package com.jiuqi.grid;

package .var.folders.cl.9twcgjfx4vsg5zl73c6wxmvm0000gn.T;

import com.jiuqi.grid.BIFF;
import com.jiuqi.grid.BitList;
import com.jiuqi.grid.CellConvertor;
import com.jiuqi.grid.CellField;
import com.jiuqi.grid.CellPostion;
import com.jiuqi.grid.CellValue;
import com.jiuqi.grid.ExtendedDatas;
import com.jiuqi.grid.GridCell;
import com.jiuqi.grid.GridCellProperty;
import com.jiuqi.grid.GridConsts;
import com.jiuqi.grid.GridFieldList;
import com.jiuqi.grid.GridFontNames;
import com.jiuqi.grid.IntList;
import com.jiuqi.grid.KeyValues;
import com.jiuqi.grid.LabelItems;
import com.jiuqi.grid.Matrix;
import com.jiuqi.grid.PropList;
import com.jiuqi.grid.ValueList;
import com.jiuqi.grid.WordList;
import com.jiuqi.util.Base64;
import com.jiuqi.util.MemStream;
import com.jiuqi.util.Stream;
import com.jiuqi.util.StreamException;
import com.jiuqi.util.StringUtils;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.Serializable;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

public final class GridData implements Serializable, Cloneable {
  private IntList cellColors;
  
  private IntList fontColors;
  
  private IntList edgeColors;
  
  private GridFontNames fontNames;
  
  private IntList fontSizes;
  
  private IntList colWidths;
  
  private IntList rowHeights;
  
  private BitList colVisible;
  
  private BitList rowVisible;
  
  private BitList hideDataRow;
  
  private BitList hideDataCol;
  
  private BitList colAutoSize;
  
  private BitList rowAutoSize;
  
  private Matrix values;
  
  private Matrix cells;
  
  private GridFieldList rects;
  
  private GridCell cell;
  
  private PropList propList;
  
  private Matrix objects;
  
  private int scrollTopCol = -1;
  
  private int scrollTopRow = -1;
  
  private int scrollBottomCol = -1;
  
  private int scrollBottomRow = -1;
  
  private int curCol = -1;
  
  private int curRow = -1;
  
  private boolean useFontSize;
  
  private LabelItems labels = new LabelItems();
  
  private Matrix cellCSS = null;
  
  private String gridDataClass;
  
  private String referenceBase;
  
  private Matrix cellScript = null;
  
  private int horzOffset;
  
  private int vertOffset;
  
  private boolean horzCenter;
  
  private boolean vertCenter;
  
  private ExtendedDatas extDatas;
  
  private int options;
  
  private Set modifiedPropIndexRecycle;
  
  private boolean modifying = false;
  
  private int colCountPerPage;
  
  private int rowCountPerPage;
  
  private boolean rowPagingFirst;
  
  public static final int GRID_AUTOMERGE = 1;
  
  public static final int GRID_AUTOSTYLE = 2;
  
  public static final int GRID_DYNAMICSTORE = 4;
  
  public static final int GRID_FULLSTYLE = 8;
  
  public GridData(int colCount, int rowCount) {
    this();
    setColCount(colCount);
    setRowCount(rowCount);
  }
  
  public GridData(int options) {
    this();
    setOptions(options);
  }
  
  public GridData(int colCount, int rowCount, int options) {
    this(colCount, rowCount);
    setOptions(options);
  }
  
  public void beginModify() {
    if (isModifying())
      throw new IllegalStateException("无法对已经处于修改状态的GridData执行beginModify方法！"); 
    if (this.modifiedPropIndexRecycle == null)
      this.modifiedPropIndexRecycle = new HashSet(); 
    this.modifying = true;
  }
  
  public void secureBeginModify() {
    synchronized (this) {
      beginModify();
    } 
  }
  
  public boolean isModifying() {
    return this.modifying;
  }
  
  public void endModify() {
    rebuildPropIndexes(this.modifiedPropIndexRecycle);
    this.modifiedPropIndexRecycle.clear();
    this.modifying = false;
  }
  
  public void secureEndModify() {
    synchronized (this) {
      endModify();
    } 
  }
  
  private void createCellCssMatrix() {
    if (this.cellCSS == null) {
      this.cellCSS = new Matrix();
      this.cellCSS.setSize(getColCount(), getRowCount());
    } 
  }
  
  private void createCellScriptMatrix() {
    if (this.cellScript == null) {
      this.cellScript = new Matrix();
      this.cellScript.setSize(getColCount(), getRowCount());
    } 
  }
  
  private boolean checkCol(int col, boolean isInsert) {
    if (isInsert)
      return (col > 0 && col <= getColCount()); 
    return (col >= 0 && col < getColCount());
  }
  
  private boolean checkRow(int row, boolean isInsert) {
    if (isInsert)
      return (row > 0 && row <= getRowCount()); 
    return (row >= 0 && row < getRowCount());
  }
  
  private void rebuildPropIndex(int col, int row, int oldIndex) {
    if (oldIndex < 0)
      return; 
    if (this.modifying) {
      this.modifiedPropIndexRecycle.add(new Integer(oldIndex));
      return;
    } 
    if (!this.cells.hasElement(new Integer(oldIndex))) {
      for (int x = 0; x < this.cells.getXCount(); x++) {
        for (int y = 0; y < this.cells.getYCount(); y++) {
          Integer index = (Integer)this.cells.getElement(x, y);
          if (index != null)
            if (index.intValue() > oldIndex) {
              index = new Integer(index.intValue() - 1);
              this.cells.setElement(x, y, index);
            }  
        } 
      } 
      this.propList.remove(oldIndex);
    } 
  }
  
  private void rebuildPropIndexes(Set bufferedIndex) {
    if (bufferedIndex.size() == 0 || this.propList.count() == 0)
      return; 
    int[] propCase = new int[this.propList.count()];
    Iterator indexIt = bufferedIndex.iterator();
    while (indexIt.hasNext()) {
      Object index = indexIt.next();
      if (!this.cells.hasElement(index))
        propCase[((Integer)index).intValue()] = -1; 
    } 
    int ci;
    for (ci = propCase.length - 1; ci >= 0; ci--) {
      if (propCase[ci] != 0)
        this.propList.remove(ci); 
    } 
    for (ci = 1; ci < propCase.length; ci++)
      propCase[ci] = propCase[ci] + propCase[ci - 1]; 
    if (propCase[propCase.length - 1] == 0)
      return; 
    for (int x = 0; x < this.cells.getXCount(); x++) {
      for (int y = 0; y < this.cells.getYCount(); y++) {
        Integer index = (Integer)this.cells.getElement(x, y);
        if (index != null)
          if (propCase[index.intValue()] != 0) {
            index = new Integer(index.intValue() + propCase[index.intValue()]);
            this.cells.setElement(x, y, index);
          }  
      } 
    } 
  }
  
  public String getDataClass() {
    return this.gridDataClass;
  }
  
  public void setDataClass(String value) {
    this.gridDataClass = value;
  }
  
  public String getReferenceBase() {
    return this.referenceBase;
  }
  
  public void setReferenceBase(String value) {
    this.referenceBase = value;
  }
  
  public String getCellData(int col, int row) {
    return (String)this.values.getElement(col, row);
  }
  
  public String getCellData(String pos) {
    CellPostion p = new CellPostion(pos);
    return getCellData(p.getCol(), p.getRow());
  }
  
  public void setCellData(int col, int row, String value) {
    this.values.setElement(col, row, value);
  }
  
  public void setCellData(String pos, String value) {
    CellPostion p = new CellPostion(pos);
    setCellData(p.getCol(), p.getRow(), value);
  }
  
  public String getCellClass(int col, int row) {
    return (this.cellCSS == null) ? null : (String)this.cellCSS.getElement(col, row);
  }
  
  public void setCellClass(int col, int row, String value) {
    if (this.cellCSS == null) {
      if (value == null || value.length() == 0)
        return; 
      createCellCssMatrix();
    } 
    this.cellCSS.setElement(col, row, value);
  }
  
  public boolean getUseFontSize() {
    return this.useFontSize;
  }
  
  public void setUserFontSize(boolean value) {
    this.useFontSize = value;
  }
  
  public String getCellScript(int col, int row) {
    return (this.cellScript == null) ? null : (String)this.cellScript.getElement(col, row);
  }
  
  public void setCellScript(int col, int row, String value) {
    if (this.cellScript == null) {
      if (value == null || value.length() == 0)
        return; 
      createCellScriptMatrix();
    } 
    this.cellScript.setElement(col, row, value);
  }
  
  public int getCellStyleIndex(int col, int row) {
    Integer index = (Integer)this.cells.getElement(col, row);
    return (index == null) ? 0 : index.intValue();
  }
  
  public GridCell[] getCellSytles() {
    if (this.propList.count() > 0) {
      GridCell[] ret = new GridCell[this.propList.count()];
      for (int i = 0; i < ret.length; i++) {
        ret[i] = new GridCell();
        ret[i].init(this, 0, 0, this.propList.get(i));
      } 
      return ret;
    } 
    return new GridCell[0];
  }
  
  public GridCell[] getCellStyles(int left, int top, int right, int bottom, boolean fullMode) {
    if (this.propList.count() == 0)
      return new GridCell[0]; 
    GridCell[] ret = new GridCell[this.propList.count()];
    for (int col = left; col <= right; col++) {
      for (int row = top; row <= bottom; row++) {
        if (StringUtils.isEmpty(getCellClass(col, row))) {
          int index = getCellStyleIndex(col, row);
          if (ret[index] == null) {
            ret[index] = new GridCell();
            if (fullMode) {
              ret[index].init(this, col, row);
            } else {
              ret[index].init(this, 0, 0, getCellPropData(col, row));
            } 
          } 
        } 
      } 
    } 
    return ret;
  }
  
  public byte[] getCellPropData(int col, int row) {
    Integer index = (Integer)this.cells.getElement(col, row);
    if (index != null)
      return this.propList.get(index.intValue()); 
    if (col == 0 || row == 0)
      return GridConsts.DEF_HEADCELL_PROP; 
    return GridConsts.DEF_CELL_PROP;
  }
  
  public GridCell getCell(int col, int row) {
    this.cell.init(this, col, row);
    return this.cell;
  }
  
  public GridCell getCell(String pos) {
    CellPostion p = new CellPostion(pos);
    return getCell(p.getCol(), p.getRow());
  }
  
  public GridCell getCellEx(int col, int row) {
    GridCell result = new GridCell();
    result.init(this, col, row);
    return result;
  }
  
  public GridCell getCellEx(String pos) {
    CellPostion p = new CellPostion(pos);
    return getCellEx(p.getCol(), p.getRow());
  }
  
  public GridCellProperty getCellForChange(int col, int row) {
    if (!checkCol(col, false) || !checkRow(row, false))
      return null; 
    GridCellProperty result = new GridCellProperty();
    result.init(this, col, row);
    return result;
  }
  
  public GridCellProperty getCellForChange(String pos) {
    CellPostion p = new CellPostion(pos);
    return getCellForChange(p.getCol(), p.getRow());
  }
  
  public void setCell(int col, int row, GridCell value) {
    int index = this.propList.indexOf(value.getPropData());
    if (index < 0)
      index = this.propList.add(value.getPropData()); 
    boolean dynamicStore = ((this.options & 0x4) != 0);
    Integer oldIndex = dynamicStore ? (Integer)this.cells.getElement(col, row) : null;
    this.cells.setElement(col, row, new Integer(index));
    if (dynamicStore && oldIndex != null && oldIndex.intValue() != index)
      rebuildPropIndex(col, row, oldIndex.intValue()); 
  }
  
  public void setCell(GridCell value) {
    setCell(value.getColNum(), value.getRowNum(), value);
  }
  
  public void setCell(int left, int top, int right, int bottom, GridCell value) {
    setCell(new CellField(left, top, right, bottom), value);
  }
  
  public void setCell(CellField field, GridCell value) {
    int index = this.propList.indexOf(value.getPropData());
    if (index == -1)
      index = this.propList.add(value.getPropData()); 
    Integer p = new Integer(index);
    for (int col = field.left; col <= field.right; col++) {
      for (int row = field.top; row <= field.bottom; row++) {
        CellField f = this.rects.getMergeRect(col, row);
        if (f == null || (f.left == col && f.top == row))
          this.cells.setElement(col, row, p); 
      } 
    } 
  }
  
  public void setCell(CellField[] fields, GridCell value) {
    if (fields == null || fields.length == 0)
      return; 
    int index = this.propList.indexOf(value.getPropData());
    if (index == -1)
      index = this.propList.add(value.getPropData()); 
    Integer p = new Integer(index);
    for (int i = 0; i < fields.length; i++) {
      for (int col = (fields[i]).left; col <= (fields[i]).right; col++) {
        for (int row = (fields[i]).top; row <= (fields[i]).bottom; row++) {
          CellField f = this.rects.getMergeRect(col, row);
          if (f == null || (f.left == col && f.top == row))
            this.cells.setElement(col, row, p); 
        } 
      } 
    } 
  }
  
  public Object getObj(int col, int row) {
    return this.objects.getElement(col, row);
  }
  
  public void setObj(int col, int row, Object obj) {
    this.objects.setElement(col, row, obj);
  }
  
  public boolean getColAutoSize(int col) {
    return this.colAutoSize.get(col);
  }
  
  public void setColAutoSize(int col, boolean value) {
    this.colAutoSize.set(col, value);
  }
  
  public boolean getColVisible(int col) {
    return (this.colVisible.count() > 0) ? this.colVisible.get(col) : true;
  }
  
  public void setColVisible(int col, boolean value) {
    if (!value || col < this.colVisible.count())
      this.colVisible.set(col, value); 
  }
  
  public int getColCount() {
    return this.colWidths.count();
  }
  
  public int getColWidths(int col) {
    return this.colWidths.get(col);
  }
  
  public void setColWidths(int col, int value) {
    this.colWidths.set(col, value);
  }
  
  public boolean getRowAutoSize(int row) {
    return this.rowAutoSize.get(row);
  }
  
  public void setRowAutoSize(int row, boolean value) {
    this.rowAutoSize.set(row, value);
  }
  
  public boolean getRowVisible(int row) {
    return (this.rowVisible.count() > 0) ? this.rowVisible.get(row) : true;
  }
  
  public void setRowVisible(int row, boolean value) {
    if (!value || row < getRowCount())
      this.rowVisible.set(row, value); 
  }
  
  public boolean getHideDataRow(int row) {
    return this.hideDataRow.get(row);
  }
  
  public void setHideDataRow(int row, boolean value) {
    if (!value || row < getRowCount())
      this.hideDataRow.set(row, value); 
  }
  
  public boolean getHideDataCol(int col) {
    return this.hideDataCol.get(col);
  }
  
  public int[][] getHideDataCols() {
    int[][] result = (int[][])null;
    int count = 0;
    boolean state = false;
    for (int i = 1; i < getColCount(); i++) {
      if (!state && getHideDataCol(i)) {
        count++;
        state = true;
      } 
      if (state && !getHideDataCol(i))
        state = false; 
    } 
    if (count > 0) {
      result = new int[2][count];
      state = false;
      int k = 0;
      for (int j = 1; j < getColCount(); j++) {
        if (!state && getHideDataCol(j)) {
          result[0][k] = j;
          state = true;
        } 
        if (state && !getHideDataCol(j)) {
          state = false;
          result[1][k] = j - 1;
          k++;
        } 
      } 
      if (state)
        result[1][count - 1] = getColCount() - 1; 
    } 
    return result;
  }
  
  public int[][] getHideDataRows() {
    int[][] result = (int[][])null;
    int count = 0;
    boolean state = false;
    for (int i = 1; i < getRowCount(); i++) {
      if (!state && getHideDataRow(i)) {
        count++;
        state = true;
      } 
      if (state && !getHideDataRow(i))
        state = false; 
    } 
    if (count > 0) {
      result = new int[2][count];
      state = false;
      int k = 0;
      for (int j = 1; j < getRowCount(); j++) {
        if (!state && getHideDataRow(j)) {
          result[0][k] = j;
          state = true;
        } 
        if (state && !getHideDataRow(j)) {
          state = false;
          result[1][k] = j - 1;
          k++;
        } 
      } 
      if (state)
        result[1][count - 1] = getRowCount() - 1; 
    } 
    return result;
  }
  
  public void setHideDataCol(int col, boolean value) {
    if (!value || col < getColCount())
      this.hideDataCol.set(col, value); 
  }
  
  public int getRowCount() {
    return this.rowHeights.count();
  }
  
  public void setColCount(int value) {
    int colCount = getColCount();
    if (value > 0 && value < colCount) {
      int delCount = colCount - value;
      this.colVisible.delete(value, delCount);
      this.colAutoSize.delete(value, delCount);
      this.colWidths.delete(value, delCount);
      this.values.xDelete(value, delCount);
      this.cells.xDelete(value, delCount);
      this.objects.xDelete(value, delCount);
      if (this.cellCSS != null)
        this.cellCSS.xDelete(value, delCount); 
      if (this.cellScript != null)
        this.cellScript.xDelete(value, delCount); 
      this.rects.delCols(value, delCount);
    } else if (value > colCount) {
      int insCount = value - colCount;
      this.colVisible.insert(colCount, insCount);
      this.colAutoSize.insert(colCount, insCount);
      this.colWidths.insert(colCount, insCount, (colCount == 0) ? 100 : this.colWidths.get(colCount - 1));
      if (colCount == 0)
        this.colWidths.set(0, 50); 
      this.values.xInsert(colCount, insCount);
      this.cells.xInsert(colCount, insCount);
      this.objects.xInsert(colCount, insCount);
      if (this.cellCSS != null)
        this.cellCSS.xInsert(colCount, insCount); 
      if (this.cellScript != null)
        this.cellScript.xInsert(colCount, insCount); 
      this.rects.insertCols(colCount, insCount);
      if (getRowCount() > 0)
        for (int i = colCount; i < colCount + insCount; i++) {
          GridCell cell = getCell(i, 0);
          setCell(i, 0, cell);
        }  
    } 
    colCount = getColCount();
    if (this.scrollTopCol >= colCount)
      setScrollTopCol(colCount - 1); 
    if (this.scrollBottomCol >= colCount)
      setScrollBottomCol(colCount - 1); 
  }
  
  public void setRowCount(int value) {
    int rowCount = getRowCount();
    if (value < rowCount) {
      int delCount = rowCount - value;
      this.rowVisible.delete(value, delCount);
      this.rowAutoSize.delete(value, delCount);
      this.rowHeights.delete(value, delCount);
      this.values.yDelete(value, delCount);
      this.cells.yDelete(value, delCount);
      this.objects.yDelete(value, delCount);
      if (this.cellCSS != null)
        this.cellCSS.yDelete(value, delCount); 
      if (this.cellScript != null)
        this.cellScript.yDelete(value, delCount); 
      this.rects.delRows(value, delCount);
    } else if (value > getRowCount()) {
      int insCount = value - rowCount;
      this.rowVisible.insert(rowCount, insCount);
      this.rowAutoSize.insert(rowCount, insCount);
      this.rowHeights.insert(rowCount, insCount, (rowCount == 0) ? 20 : this.rowHeights.get(rowCount - 1));
      this.values.yInsert(rowCount, insCount);
      this.objects.yInsert(rowCount, insCount);
      this.cells.yInsert(rowCount, insCount);
      if (this.cellCSS != null)
        this.cellCSS.yInsert(rowCount, insCount); 
      if (this.cellScript != null)
        this.cellScript.yInsert(rowCount, insCount); 
      this.rects.insertRows(rowCount, insCount);
      if (getColCount() > 0)
        for (int i = rowCount; i < rowCount + insCount; i++) {
          GridCell cell = getCell(0, i);
          setCell(0, i, cell);
        }  
    } 
  }
  
  public void insertCol(int index, int count) {
    insertCol(index, count, ((this.options & 0x2) == 0), ((this.options & 0x8) != 0));
  }
  
  private void insertCol(int index, int count, boolean fastMode, boolean fullMode) {
    if (!checkCol(index, true))
      return; 
    this.colVisible.insert(index, count);
    if (index > 0) {
      this.colAutoSize.insert(index, count, this.colAutoSize.get(index - 1));
    } else {
      this.colAutoSize.insert(index, count);
    } 
    this.colWidths.insert(index, count, (getColCount() == 0 || index == 1) ? 100 : this.colWidths.get(index - 1));
    this.values.xInsert(index, count);
    this.cells.xInsert(index, count);
    this.objects.xInsert(index, count);
    if (this.cellCSS != null)
      this.cellCSS.xInsert(index, count); 
    if (this.cellScript != null)
      this.cellScript.xInsert(index, count); 
    this.rects.insertCols(index, count);
    if (getRowCount() > 0 && !fastMode)
      for (int row = 1; row < getRowCount(); row++) {
        GridCell cell;
        if (index > 1) {
          cell = getCell(index - 1, row);
        } else {
          cell = this.cell;
          cell.init(this, -1, -1, GridConsts.DEF_CELL_PROP);
        } 
        for (int col = index; col < index + count; col++) {
          setCell(col, row, cell);
          if (fullMode) {
            setCellClass(col, row, cell.getCssClass());
            setCellScript(col, row, cell.getScript());
          } 
        } 
      }  
    if (this.scrollTopCol >= index)
      this.scrollTopCol += count; 
    if (this.scrollBottomCol + 1 >= index)
      this.scrollBottomCol += count; 
  }
  
  public void deleteCol(int index, int count) {
    if (!checkCol(index, false))
      return; 
    if (index + count > getColCount())
      count = getColCount() - index; 
    this.colVisible.delete(index, count);
    this.colAutoSize.delete(index, count);
    this.colWidths.delete(index, count);
    this.values.xDelete(index, count);
    this.cells.xDelete(index, count);
    this.objects.xDelete(index, count);
    if (this.cellCSS != null)
      this.cellCSS.xDelete(index, count); 
    if (this.cellScript != null)
      this.cellScript.xDelete(index, count); 
    this.rects.delCols(index, count);
    if (this.scrollTopCol >= index)
      this.scrollTopCol = 
        Math.min(Math.max(index, this.scrollTopCol - count), getColCount() - 1); 
    if (this.scrollBottomCol >= index)
      this.scrollBottomCol = 
        Math.min(Math.max(index, this.scrollBottomCol - count), getColCount() - 1); 
  }
  
  public void insertRow(int index, int count) {
    insertRow(index, count, ((this.options & 0x2) == 0), ((this.options & 0x8) != 0));
  }
  
  private void insertRow(int index, int count, boolean fastMode, boolean fullMode) {
    if (!checkRow(index, true))
      return; 
    this.rowVisible.insert(index, count);
    if (index > 0) {
      this.rowAutoSize.insert(index, count, this.rowAutoSize.get(index - 1));
    } else {
      this.rowAutoSize.insert(index, count);
    } 
    this.rowHeights.insert(index, count, (getRowCount() == 0 || index == 1) ? 20 : this.rowHeights.get(index - 1));
    this.values.yInsert(index, count);
    this.cells.yInsert(index, count);
    this.objects.yInsert(index, count);
    if (this.cellCSS != null)
      this.cellCSS.yInsert(index, count); 
    if (this.cellScript != null)
      this.cellScript.yInsert(index, count); 
    this.rects.insertRows(index, count);
    if (getColCount() > 0 && !fastMode)
      for (int col = 1; col < getColCount(); col++) {
        GridCell cell;
        if (index > 1) {
          cell = getCell(col, index - 1);
        } else {
          cell = this.cell;
          cell.init(this, -1, -1, GridConsts.DEF_CELL_PROP);
        } 
        for (int row = index; row < index + count; row++) {
          setCell(col, row, cell);
          if (fullMode) {
            setCellClass(col, row, cell.getCssClass());
            setCellScript(col, row, cell.getScript());
          } 
        } 
      }  
    if (this.scrollTopRow >= index)
      this.scrollTopRow += count; 
    if (this.scrollBottomRow + 1 >= index)
      this.scrollBottomRow += count; 
  }
  
  public boolean duplicateCols(int fromIndex, int toIndex, int duplicateCount, int insertIndex) {
    if (!checkCol(fromIndex, false))
      return false; 
    if (!checkCol(toIndex, false))
      return false; 
    if (!checkCol(insertIndex, true))
      return false; 
    if (toIndex < fromIndex || duplicateCount <= 0 || (insertIndex > fromIndex && insertIndex <= toIndex))
      return false; 
    int insertCount = (toIndex - fromIndex + 1) * duplicateCount;
    insertCol(insertIndex, insertCount, true, true);
    if (insertIndex <= fromIndex) {
      fromIndex += insertCount;
      toIndex += insertCount;
    } 
    List<CellField> merges = (fromIndex < toIndex) ? this.rects.findFieldsInCols(fromIndex, toIndex) : null;
    for (int i = 0; i < duplicateCount; i++) {
      int offset = insertIndex + (toIndex - fromIndex + 1) * i - fromIndex;
      int j;
      for (j = fromIndex; j <= toIndex; j++) {
        this.colWidths.set(offset + j, this.colWidths.get(j));
        this.values.xCopy(j, offset + j);
        if (this.cellCSS != null)
          this.cellCSS.xCopy(j, offset + j); 
        if (this.cellScript != null)
          this.cellScript.xCopy(j, offset + j); 
        this.cells.xCopy(j, offset + j);
        if (!this.colVisible.get(j))
          this.colVisible.set(j + offset, false); 
      } 
      if (merges != null && !merges.isEmpty())
        for (j = 0; j < merges.size(); j++) {
          CellField cf = merges.get(j);
          mergeCells(cf.left + offset, cf.top, cf.right + offset, cf.bottom);
        }  
    } 
    return true;
  }
  
  public boolean duplicateRows(int fromIndex, int toIndex, int duplicateCount, int insertIndex) {
    if (!checkRow(fromIndex, false))
      return false; 
    if (!checkRow(toIndex, false))
      return false; 
    if (!checkRow(insertIndex, true))
      return false; 
    if (toIndex < fromIndex || duplicateCount <= 0 || (insertIndex > fromIndex && insertIndex <= toIndex))
      return false; 
    int insertCount = (toIndex - fromIndex + 1) * duplicateCount;
    insertRow(insertIndex, insertCount, true, true);
    if (insertIndex <= fromIndex) {
      fromIndex += insertCount;
      toIndex += insertCount;
    } 
    List<CellField> merges = (fromIndex < toIndex) ? this.rects.findFieldsInRows(fromIndex, toIndex) : null;
    for (int i = 0; i < duplicateCount; i++) {
      int offset = insertIndex + (toIndex - fromIndex + 1) * i - fromIndex;
      int j;
      for (j = fromIndex; j <= toIndex; j++) {
        this.rowHeights.set(offset + j, this.rowHeights.get(j));
        this.values.yCopy(j, offset + j);
        if (this.cellCSS != null)
          this.cellCSS.yCopy(j, offset + j); 
        if (this.cellScript != null)
          this.cellScript.yCopy(j, offset + j); 
        this.cells.yCopy(j, offset + j);
        if (!this.rowVisible.get(j))
          this.rowVisible.set(j + offset, false); 
      } 
      if (merges != null && !merges.isEmpty())
        for (j = 0; j < merges.size(); j++) {
          CellField cf = merges.get(j);
          mergeCells(cf.left, cf.top + offset, cf.right, cf.bottom + offset);
        }  
    } 
    return true;
  }
  
  public void deleteRow(int index, int count) {
    if (!checkRow(index, false))
      return; 
    if (index + count > getRowCount())
      count = getRowCount() - index; 
    this.rowVisible.delete(index, count);
    this.rowAutoSize.delete(index, count);
    this.rowHeights.delete(index, count);
    this.values.yDelete(index, count);
    this.cells.yDelete(index, count);
    if (this.cellCSS != null)
      this.cellCSS.yDelete(index, count); 
    if (this.cellScript != null)
      this.cellScript.yDelete(index, count); 
    this.objects.yDelete(index, count);
    this.rects.delRows(index, count);
    if (this.scrollTopRow >= index)
      this.scrollTopRow = Math.min(Math.max(index, this.scrollTopRow - count), getRowCount() - 1); 
    if (this.scrollBottomRow >= index)
      this.scrollBottomRow = Math.min(Math.max(index, this.scrollBottomRow - count), getRowCount() - 1); 
  }
  
  public int getRowHeights(int row) {
    return this.rowHeights.get(row);
  }
  
  public void setRowHeights(int row, int value) {
    this.rowHeights.set(row, value);
  }
  
  public int getScrollTopCol() {
    return this.scrollTopCol;
  }
  
  public void setScrollTopCol(int value) {
    this.scrollTopCol = value;
  }
  
  public int getScrollTopRow() {
    return this.scrollTopRow;
  }
  
  public void setScrollTopRow(int value) {
    this.scrollTopRow = value;
  }
  
  public int getScrollBottomCol() {
    return this.scrollBottomCol;
  }
  
  public void setScrollBottomCol(int value) {
    this.scrollBottomCol = value;
  }
  
  public int getScrollBottomRow() {
    return this.scrollBottomRow;
  }
  
  public void setScrollBottomRow(int value) {
    this.scrollBottomRow = value;
  }
  
  public int lockTopCol() {
    return this.scrollTopCol;
  }
  
  public void setLockTopCol(int value) {
    this.scrollTopCol = value;
  }
  
  public int lockTopRow() {
    return this.scrollTopRow;
  }
  
  public void setLockTopRow(int value) {
    this.scrollTopRow = value;
  }
  
  public int lockBottomCol() {
    return this.scrollBottomCol;
  }
  
  public void setLockBottomCol(int value) {
    this.scrollBottomCol = value;
  }
  
  public int lockBottomRow() {
    return this.scrollBottomRow;
  }
  
  public void setLockBottomRow(int value) {
    this.scrollBottomRow = value;
  }
  
  public int curCol() {
    return this.curCol;
  }
  
  public void setCurCol(int value) {
    this.curCol = value;
  }
  
  public int curRow() {
    return this.curRow;
  }
  
  public void setCurRow(int value) {
    this.curRow = value;
  }
  
  public GridFieldList merges() {
    return this.rects;
  }
  
  public IntList cellColors() {
    return this.cellColors;
  }
  
  public IntList fontColors() {
    return this.fontColors;
  }
  
  public IntList edgeColors() {
    return this.edgeColors;
  }
  
  public void setEdgeColor(int index, int color) {
    if (index >= 0 && index < 16)
      this.edgeColors.set(index, color); 
  }
  
  public GridFontNames fontNames() {
    return this.fontNames;
  }
  
  public IntList fontSizes() {
    return this.fontSizes;
  }
  
  public boolean cellMerged(int col, int row) {
    return (this.rects.getMergeRect(col, row) != null);
  }
  
  public CellField expandCell(int col, int row) {
    CellField rt = this.rects.getMergeRect(col, row);
    return (rt == null) ? new CellField(col, row, col, row) : (CellField)rt.clone();
  }
  
  public boolean mergeCells(int left, int top, int right, int bottom) {
    CellField cf = new CellField(left, top, right, bottom);
    return this.rects.addMergeRect(cf);
  }
  
  public LabelItems getLabels() {
    return this.labels;
  }
  
  private void clear() {
    this.scrollTopCol = -1;
    this.scrollTopRow = -1;
    this.scrollBottomCol = -1;
    this.scrollBottomRow = -1;
    this.curCol = -1;
    this.curRow = -1;
    this.labels.clear();
    this.extDatas.clear();
    this.cellCSS = null;
    this.cellScript = null;
  }
  
  public void loadFromStream(Stream stream) throws StreamException {
    int fileVer;
    ValueList valueList = new ValueList();
    ValueList cellCSSList = new ValueList();
    ValueList cellScriptList = new ValueList();
    WordList propIndexs = new WordList();
    KeyValues dataProps = new KeyValues();
    BIFF b = new BIFF();
    b.data().setUseEncode(stream.getUseEncode());
    b.data().setCharset(stream.getCharset());
    long p = stream.getPosition();
    BIFF.readBIFF(1, b, stream);
    String signFlag = b.data().readString(4);
    if (!"BIFF".equals(signFlag)) {
      stream.setPosition(p);
      BIFF.readBIFF(2, b, stream);
      signFlag = b.data().readString(4);
      if (!"BIFF".equals(signFlag))
        throw new StreamException("错误的文件格式"); 
    } 
    int ver = b.data().readInt();
    if (b.data().getPosition() < b.data().getSize()) {
      fileVer = b.data().readInt();
    } else {
      fileVer = 16777216;
    } 
    clear();
    this.useFontSize = false;
    while (b.ident != 127) {
      BIFF.readBIFF(ver, b, stream);
      switch (b.ident) {
        case 25:
          dataProps.loadFromStream(b.data());
        case 18:
          this.useFontSize = true;
        case 6:
          this.fontNames.loadFromStream(b.data());
        case 8:
          this.fontSizes.loadFromStream(b.data(), b.size);
        case 7:
          this.fontColors.loadFromStream(b.data(), b.size);
        case 10:
          this.cellColors.loadFromStream(b.data(), b.size);
        case 9:
          this.edgeColors.loadFromStream(b.data(), b.size);
        case 2:
          this.colWidths.loadFromStream(b.data(), b.size);
        case 13:
          this.colAutoSize.loadFromStream(b.data(), b.size);
        case 12:
          this.colVisible.loadFromStream(b.data(), b.size);
        case 21:
          this.hideDataCol.loadFromStream(b.data(), b.size);
        case 22:
          this.hideDataRow.loadFromStream(b.data(), b.size);
        case 3:
          this.rowHeights.loadFromStream(b.data(), b.size);
        case 15:
          this.rowAutoSize.loadFromStream(b.data(), b.size);
        case 14:
          this.rowVisible.loadFromStream(b.data(), b.size);
        case 4:
          valueList.loadFromStream(b.data());
        case 11:
          this.propList.loadFromStreamEx(b.data(), 0, 12);
        case 19:
          this.propList.loadFromStreamEx(b.data(), 12, 8);
        case 1:
          propIndexs.loadFromStream(b.data(), b.size);
        case 5:
          this.rects.loadFromStream(b.data(), b.size);
        case 16:
          this.scrollTopCol = b.data().readInt();
          this.scrollTopRow = b.data().readInt();
          this.scrollBottomCol = b.data().readInt();
          this.scrollBottomRow = b.data().readInt();
        case 17:
          this.curCol = b.data().readInt();
          this.curRow = b.data().readInt();
        case 20:
          this.labels.loadFromStream(b.data());
          if (fileVer == 16777216)
            this.labels.ver10Tover11(); 
        case 23:
          this.colCountPerPage = b.data().readInt();
          this.rowCountPerPage = b.data().readInt();
          this.rowPagingFirst = b.data().readBool();
        case 24:
          cellCSSList.loadFromStream(b.data());
        case 26:
          cellScriptList.loadFromStream(b.data());
        case 27:
          this.extDatas.loadFromStream(b.data(), b.size);
      } 
    } 
    this.gridDataClass = dataProps.getValue("gridclass");
    this.referenceBase = dataProps.getValue("gridrefbase");
    this.horzOffset = dataProps.getAsInt("horzoffset");
    this.vertOffset = dataProps.getAsInt("vertoffset");
    this.horzCenter = dataProps.getAsBoolean("horzcenter");
    this.vertCenter = dataProps.getAsBoolean("vertcenter");
    setColCount(this.colWidths.count());
    setRowCount(this.rowHeights.count());
    this.cells.resetSize(this.colWidths.count(), this.rowHeights.count());
    this.objects.resetSize(this.colWidths.count(), this.rowHeights.count());
    this.values.resetSize(this.colWidths.count(), this.rowHeights.count());
    int i = 0;
    for (int c = 0; c < this.colWidths.count(); c++) {
      for (int r = 0; r < this.rowHeights.count(); r++) {
        this.cells.setElement(c, r, new Integer(propIndexs.get(i)));
        i++;
      } 
    } 
    for (i = 0; i < valueList.count(); i++)
      this.values.setElement(valueList.col(i), valueList.row(i), valueList.get(i)); 
    for (i = 0; i < cellCSSList.count(); i++)
      setCellClass(cellCSSList.col(i), cellCSSList.row(i), cellCSSList.get(i)); 
    for (i = 0; i < cellScriptList.count(); i++)
      setCellScript(cellScriptList.col(i), cellScriptList.row(i), cellScriptList.get(i)); 
  }
  
  public void saveToStream(Stream stream) throws StreamException {
    ValueList valueList = new ValueList();
    ValueList cellCSSList = new ValueList();
    ValueList cellScriptList = new ValueList();
    WordList propIndexs = new WordList();
    KeyValues dataProps = new KeyValues();
    dataProps.setValue("gridclass", this.gridDataClass);
    dataProps.setValue("gridrefbase", this.referenceBase);
    dataProps.setAsInt("horzoffset", this.horzOffset);
    dataProps.setAsInt("vertoffset", this.vertOffset);
    dataProps.setAsBoolean("horzcenter", this.horzCenter);
    dataProps.setAsBoolean("vertcenter", this.vertCenter);
    for (int c = 0; c < this.colWidths.count(); c++) {
      for (int r = 0; r < this.rowHeights.count(); r++) {
        String cell = (String)this.values.getElement(c, r);
        if (cell != null && cell.length() > 0) {
          CellValue cv = new CellValue();
          cv.col = c;
          cv.row = r;
          cv.value = cell;
          valueList.add(cv);
        } 
        Integer propIndex = (Integer)this.cells.getElement(c, r);
        if (propIndex == null) {
          int pi = this.propList.indexOf((c == 0 || r == 0) ? GridConsts.DEF_HEADCELL_PROP : GridConsts.DEF_CELL_PROP);
          if (c == 0 || r == 0) {
            pi = this.propList.add(GridConsts.DEF_HEADCELL_PROP);
          } else {
            pi = this.propList.add(GridConsts.DEF_CELL_PROP);
          } 
          propIndex = new Integer(pi);
        } 
        propIndexs.add(propIndex);
        cell = getCellClass(c, r);
        if (cell != null && cell.length() > 0) {
          CellValue cv = new CellValue();
          cv.col = c;
          cv.row = r;
          cv.value = cell;
          cellCSSList.add(cv);
        } 
        cell = getCellScript(c, r);
        if (cell != null && cell.length() != 0) {
          CellValue cv = new CellValue();
          cv.col = c;
          cv.row = r;
          cv.value = cell;
          cellScriptList.add(cv);
        } 
      } 
    } 
    BIFF b = new BIFF();
    b.ident = 0;
    b.data().writeString("BIFF");
    b.data().writeInt(2);
    b.data().writeInt(16842752);
    BIFF.writeBIFF(1, b, stream);
    b.reset();
    b.ident = 25;
    dataProps.saveToStream(b.data());
    BIFF.writeBIFF(25, b, stream);
    b.reset();
    b.ident = 18;
    BIFF.writeBIFF(2, b, stream);
    b.reset();
    b.ident = 6;
    this.fontNames.saveToStream(b.data());
    BIFF.writeBIFF(2, b, stream);
    b.reset();
    b.ident = 8;
    this.fontSizes.saveToStream(b.data());
    BIFF.writeBIFF(2, b, stream);
    b.reset();
    b.ident = 7;
    this.fontColors.saveToStream(b.data());
    BIFF.writeBIFF(2, b, stream);
    b.reset();
    b.ident = 10;
    this.cellColors.saveToStream(b.data());
    BIFF.writeBIFF(2, b, stream);
    b.reset();
    b.ident = 9;
    this.edgeColors.saveToStream(b.data());
    BIFF.writeBIFF(2, b, stream);
    b.reset();
    b.ident = 2;
    this.colWidths.saveToStream(b.data());
    BIFF.writeBIFF(2, b, stream);
    b.reset();
    b.ident = 13;
    this.colAutoSize.saveToStream(b.data());
    BIFF.writeBIFF(2, b, stream);
    b.reset();
    b.ident = 12;
    this.colVisible.saveToStream(b.data());
    BIFF.writeBIFF(2, b, stream);
    b.reset();
    b.ident = 21;
    this.hideDataCol.saveToStream(b.data());
    BIFF.writeBIFF(21, b, stream);
    b.reset();
    b.ident = 3;
    this.rowHeights.saveToStream(b.data());
    BIFF.writeBIFF(2, b, stream);
    b.reset();
    b.ident = 15;
    this.rowAutoSize.saveToStream(b.data());
    BIFF.writeBIFF(2, b, stream);
    b.reset();
    b.ident = 14;
    this.rowVisible.saveToStream(b.data());
    BIFF.writeBIFF(2, b, stream);
    b.reset();
    b.ident = 22;
    this.hideDataRow.saveToStream(b.data());
    BIFF.writeBIFF(22, b, stream);
    b.reset();
    b.ident = 4;
    valueList.saveToStream(b.data());
    BIFF.writeBIFF(2, b, stream);
    b.reset();
    b.ident = 11;
    this.propList.saveToStreamEx(b.data(), 0, 12);
    BIFF.writeBIFF(2, b, stream);
    b.reset();
    b.ident = 19;
    this.propList.saveToStreamEx(b.data(), 12, 8);
    BIFF.writeBIFF(2, b, stream);
    b.reset();
    b.ident = 1;
    propIndexs.saveToStream(b.data());
    BIFF.writeBIFF(2, b, stream);
    b.reset();
    b.ident = 5;
    this.rects.saveToStream(b.data());
    BIFF.writeBIFF(2, b, stream);
    b.reset();
    b.ident = 16;
    b.data().writeInt(this.scrollTopCol);
    b.data().writeInt(this.scrollTopRow);
    b.data().writeInt(this.scrollBottomCol);
    b.data().writeInt(this.scrollBottomRow);
    BIFF.writeBIFF(2, b, stream);
    b.reset();
    b.ident = 17;
    b.data().writeInt(this.curCol);
    b.data().writeInt(this.curRow);
    BIFF.writeBIFF(2, b, stream);
    b.reset();
    b.ident = 20;
    this.labels.saveToStream(b.data());
    BIFF.writeBIFF(2, b, stream);
    b.reset();
    b.ident = 23;
    b.data().writeInt(this.colCountPerPage);
    b.data().writeInt(this.rowCountPerPage);
    b.data().writeBool(this.rowPagingFirst);
    BIFF.writeBIFF(2, b, stream);
    b.reset();
    b.ident = 24;
    cellCSSList.saveToStream(b.data());
    BIFF.writeBIFF(2, b, stream);
    b.reset();
    b.ident = 26;
    cellScriptList.saveToStream(b.data());
    BIFF.writeBIFF(2, b, stream);
    b.reset();
    b.ident = 27;
    this.extDatas.saveToStream(b.data());
    BIFF.writeBIFF(2, b, stream);
    b.reset();
    b.ident = 127;
    BIFF.writeBIFF(2, b, stream);
  }
  
  public void loadFromStream(InputStream inStream) throws StreamException {
    MemStream stream = new MemStream();
    try {
      stream.loadFromStream(inStream);
    } catch (IOException e) {
      throw new StreamException(e);
    } 
    stream.setPosition(0L);
    loadFromStream((Stream)stream);
  }
  
  public void saveToStream(OutputStream outStream) throws StreamException {
    MemStream stream = new MemStream();
    saveToStream((Stream)stream);
    try {
      stream.saveToStream(outStream);
    } catch (IOException e) {
      throw new StreamException(e);
    } 
  }
  
  public void loadFromFile(String fileName) throws StreamException {
    MemStream stream = new MemStream();
    try {
      stream.loadFromFile(fileName);
    } catch (IOException e) {
      throw new StreamException(e);
    } 
    stream.setPosition(0L);
    loadFromStream((Stream)stream);
  }
  
  public void saveToFile(String fileName) throws StreamException {
    MemStream stream = new MemStream();
    saveToStream((Stream)stream);
    try {
      stream.saveToFile(fileName);
    } catch (IOException e) {
      throw new StreamException(e);
    } 
  }
  
  public com.jiuqi.grid.GridData makeCopy() throws StreamException {
    MemStream memStream = new MemStream();
    saveToStream((Stream)memStream);
    memStream.setPosition(0L);
    com.jiuqi.grid.GridData gd = new com.jiuqi.grid.GridData();
    gd.loadFromStream((Stream)memStream);
    gd.setOptions(getOptions());
    return gd;
  }
  
  public Object clone() {
    try {
      return makeCopy();
    } catch (StreamException streamException) {
      return null;
    } 
  }
  
  public boolean copyFrom(com.jiuqi.grid.GridData srcData, int fromLeft, int fromTop, int fromRight, int fromBottom, int toLeft, int toTop) {
    CellField dest = new CellField(toLeft, toTop, toLeft + fromRight - fromLeft, toTop + fromBottom - fromTop);
    if (!checkCol(dest.left, false))
      return false; 
    if (!checkCol(dest.right, false))
      return false; 
    if (!checkRow(dest.top, false))
      return false; 
    if (!checkRow(dest.bottom, false))
      return false; 
    if (!srcData.checkCol(fromLeft, false))
      return false; 
    if (!srcData.checkCol(fromRight, false))
      return false; 
    if (!srcData.checkRow(fromTop, false))
      return false; 
    if (!srcData.checkRow(fromBottom, false))
      return false; 
    this.rects.removeMergeRect(dest);
    GridCopyConvertor convertor = (this == srcData) ? null : new GridCopyConvertor(this, srcData);
    int x;
    for (x = fromLeft; x <= fromRight; x++) {
      for (int y = fromTop; y <= fromBottom; y++) {
        CellField field = srcData.expandCell(x, y);
        String text = srcData.getCellData(field.left, field.top);
        GridCell cell = srcData.getCellEx(field.left, field.top);
        if (field.left < fromLeft)
          field.left = fromLeft; 
        if (field.right > fromRight)
          field.right = fromRight; 
        if (field.top < fromTop)
          field.top = fromTop; 
        if (field.bottom > fromBottom)
          field.bottom = fromBottom; 
        if (field.left >= x && field.top >= y) {
          int xx = toLeft + x - fromLeft;
          int yy = toTop + y - fromTop;
          if (convertor != null)
            convertor.convert(cell); 
          setCell(xx, yy, cell);
          setCellData(xx, yy, text);
          setObj(xx, yy, srcData.getObj(field.left, field.top));
          setCellClass(xx, yy, srcData.getCellClass(field.left, field.top));
          setCellScript(xx, yy, srcData.getCellScript(field.left, field.top));
          mergeCells(xx, yy, toLeft + field.right - fromLeft, toTop + field.bottom - fromTop);
        } 
      } 
    } 
    if (dest.top == 1 && dest.bottom == getRowCount() - 1)
      for (x = fromLeft; x <= fromRight; x++) {
        int xx = toLeft + x - fromLeft;
        setHideDataCol(xx, srcData.getHideDataCol(x));
        setColAutoSize(xx, srcData.getColAutoSize(x));
        setColWidths(xx, srcData.getColWidths(x));
        setColVisible(xx, srcData.getColVisible(x));
      }  
    if (dest.left == 1 && dest.right == getColCount() - 1)
      for (int y = fromTop; y <= fromBottom; y++) {
        int yy = toTop + y - fromTop;
        setHideDataRow(yy, srcData.getHideDataRow(y));
        setRowAutoSize(yy, srcData.getRowAutoSize(y));
        setRowHeights(yy, srcData.getRowHeights(y));
        setRowVisible(yy, srcData.getRowVisible(y));
      }  
    return true;
  }
  
  public GridData() {
    this.colCountPerPage = 0;
    this.rowCountPerPage = 0;
    this.rowPagingFirst = true;
    this.cellColors = GridConsts.create64ColorsPallette();
    this.fontColors = GridConsts.create64ColorsPallette();
    this.edgeColors = GridConsts.create16ColorsPallette();
    this.fontNames = new GridFontNames(64);
    this.fontNames.addFont("宋体");
    this.fontSizes = new IntList();
    this.fontSizes.add(9);
    this.colVisible = new BitList(true);
    this.rowVisible = new BitList(true);
    this.colAutoSize = new BitList(false);
    this.rowAutoSize = new BitList(false);
    this.hideDataRow = new BitList(false);
    this.hideDataCol = new BitList(false);
    this.colWidths = new IntList();
    this.rowHeights = new IntList();
    this.values = new Matrix();
    this.cells = new Matrix();
    this.objects = new Matrix();
    this.rects = new GridFieldList(this);
    this.cell = new GridCell();
    this.propList = new PropList(20);
    this.propList.add(GridConsts.DEF_CELL_PROP);
    this.extDatas = new ExtendedDatas();
    setColCount(5);
    setRowCount(5);
    int i;
    for (i = 0; i < 5; i++) {
      GridCell cell = getCell(i, 0);
      setCell(i, 0, cell);
    } 
    for (i = 0; i < 5; i++) {
      GridCell cell = getCell(0, i);
      setCell(0, i, cell);
    } 
    this.options = 7;
    setOptions(this.options);
  }
  
  public int getColCountPerPage() {
    return this.colCountPerPage;
  }
  
  public void setColCountPerPage(int value) {
    this.colCountPerPage = value;
  }
  
  public int getRowCountPerPage() {
    return this.rowCountPerPage;
  }
  
  public void setRowCountPerPage(int value) {
    this.rowCountPerPage = value;
  }
  
  public boolean getRowPagingFirst() {
    return this.rowPagingFirst;
  }
  
  public void setRowPagingFirst(boolean value) {
    this.rowPagingFirst = value;
  }
  
  public int getColPageCount() {
    if (this.colCountPerPage <= 0)
      return 1; 
    int colCount = getColCount() - 1;
    if (this.scrollBottomCol > 0 && this.scrollBottomCol > this.scrollTopCol)
      colCount = this.scrollBottomCol; 
    if (this.scrollTopCol > 0)
      colCount -= this.scrollTopCol - 1; 
    return (colCount - 1) / this.colCountPerPage + 1;
  }
  
  public int getRowPageCount() {
    if (this.rowCountPerPage <= 0)
      return 1; 
    int rowCount = getRowCount() - 1;
    if (this.scrollBottomRow > 0 && this.scrollBottomRow > this.scrollTopRow)
      rowCount = this.scrollBottomRow; 
    if (this.scrollTopRow > 0)
      rowCount -= this.scrollTopRow - 1; 
    return (rowCount - 1) / this.rowCountPerPage + 1;
  }
  
  public int getPageCount() {
    return getColPageCount() * getRowPageCount();
  }
  
  public com.jiuqi.grid.GridData getPagedGridData(int pageIndex) {
    int px = 0;
    int py = 0;
    if (this.rowPagingFirst) {
      int count = getRowPageCount();
      px = pageIndex / count;
      py = pageIndex % count;
    } else {
      int count = getColPageCount();
      px = pageIndex % count;
      py = pageIndex / count;
    } 
    return getPagedGridData(px, py);
  }
  
  public com.jiuqi.grid.GridData getPagedGridData(int colPageIndex, int rowPageIndex) {
    if (colPageIndex < 0 || rowPageIndex < 0)
      return null; 
    int colHead = this.scrollTopCol - 1, colTail = this.scrollBottomCol + 1;
    int rowHead = this.scrollTopRow - 1, rowTail = this.scrollBottomRow + 1;
    if (colHead < 0)
      colHead = 0; 
    if (colTail <= colHead)
      colTail = getColCount(); 
    if (rowHead < 0)
      rowHead = 0; 
    if (rowTail <= rowHead)
      rowTail = getRowCount(); 
    int colCountPerPage = this.colCountPerPage;
    if (colCountPerPage <= 0)
      colCountPerPage = colTail - colHead - 1; 
    int rowCountPerPage = this.rowCountPerPage;
    if (rowCountPerPage <= 0)
      rowCountPerPage = rowTail - rowHead - 1; 
    int dataColCount = Math.min(colCountPerPage, colTail - colPageIndex * colCountPerPage - colHead - 1);
    if (dataColCount <= 0 && colPageIndex != 0)
      return null; 
    int fromDataCol = colHead + 1 + colPageIndex * colCountPerPage;
    int dataRowCount = Math.min(rowCountPerPage, rowTail - rowPageIndex * rowCountPerPage - rowHead - 1);
    if (dataRowCount <= 0 && rowPageIndex != 0)
      return null; 
    int fromDataRow = rowHead + 1 + rowPageIndex * rowCountPerPage;
    if (1 + colHead + dataColCount == colTail && 1 + rowHead + dataRowCount == rowTail)
      return this; 
    com.jiuqi.grid.GridData result = new com.jiuqi.grid.GridData(getOptions());
    result.cellColors = (IntList)this.cellColors.clone();
    result.edgeColors = (IntList)this.edgeColors.clone();
    result.fontNames = (GridFontNames)this.fontNames.clone();
    result.fontColors = (IntList)this.fontColors.clone();
    result.fontSizes = (IntList)this.fontSizes.clone();
    result.setDataClass(getDataClass());
    result.setColCount(1 + colHead + dataColCount + getColCount() - colTail);
    result.setRowCount(1 + rowHead + dataRowCount + getRowCount() - rowTail);
    result.setLockTopCol(colHead + 1);
    result.setLockTopRow(rowHead + 1);
    result.setLockBottomCol(colHead + dataColCount);
    result.setLockBottomRow(rowHead + dataRowCount);
    int i;
    for (i = 1; i <= colHead; i++) {
      result.setColWidths(i, getColWidths(i));
      result.setColAutoSize(i, getColAutoSize(i));
      result.setColVisible(i, getColVisible(i));
      result.setHideDataCol(i, getHideDataCol(i));
    } 
    for (i = 0; i < dataColCount; i++) {
      result.setColWidths(colHead + 1 + i, getColWidths(fromDataCol + i));
      result.setColAutoSize(colHead + 1 + i, getColAutoSize(fromDataCol + i));
      result.setColVisible(colHead + 1 + i, getColVisible(fromDataCol + i));
      result.setHideDataCol(colHead + 1 + i, getHideDataCol(fromDataCol + i));
    } 
    for (i = colTail; i < getColCount(); i++) {
      result.setColWidths(colHead + 1 + dataColCount + i - colTail, getColWidths(i));
      result.setColAutoSize(colHead + 1 + dataColCount + i - colTail, getColAutoSize(i));
      result.setColVisible(colHead + 1 + dataColCount + i - colTail, getColVisible(i));
      result.setHideDataCol(colHead + 1 + dataColCount + i - colTail, getHideDataCol(i));
    } 
    for (i = 1; i <= rowHead; i++) {
      result.setRowHeights(i, getRowHeights(i));
      result.setRowAutoSize(i, getRowAutoSize(i));
      result.setRowVisible(i, getRowVisible(i));
      result.setHideDataRow(i, getHideDataRow(i));
    } 
    for (i = 0; i < dataRowCount; i++) {
      result.setRowHeights(rowHead + 1 + i, getRowHeights(fromDataRow + i));
      result.setRowAutoSize(rowHead + 1 + i, getRowAutoSize(fromDataRow + i));
      result.setRowVisible(rowHead + 1 + i, getRowVisible(fromDataRow + i));
      result.setHideDataRow(rowHead + 1 + i, getHideDataRow(fromDataRow + i));
    } 
    for (i = rowTail; i < getRowCount(); i++) {
      result.setRowHeights(rowHead + 1 + dataRowCount + i - rowTail, getRowHeights(i));
      result.setRowAutoSize(rowHead + 1 + dataRowCount + i - rowTail, getRowAutoSize(i));
      result.setRowVisible(rowHead + 1 + dataRowCount + i - rowTail, getRowVisible(i));
      result.setHideDataRow(rowHead + 1 + dataRowCount + i - rowTail, getHideDataRow(i));
    } 
    if (colHead > 0 && rowHead > 0)
      result.copyFrom(this, 1, 1, colHead, rowHead, 1, 1); 
    if (dataColCount > 0 && rowHead > 0)
      result.copyFrom(this, fromDataCol, 1, fromDataCol + dataColCount - 1, rowHead, colHead + 1, 1); 
    if (colTail < getColCount() && rowHead > 0)
      result.copyFrom(this, colTail, 1, getColCount() - 1, rowHead, colHead + 1 + dataColCount, 1); 
    if (colHead > 0 && dataRowCount > 0)
      result.copyFrom(this, 1, fromDataRow, colHead, fromDataRow + dataRowCount - 1, 1, rowHead + 1); 
    if (dataColCount > 0 && dataRowCount > 0)
      result.copyFrom(this, fromDataCol, fromDataRow, fromDataCol + dataColCount - 1, fromDataRow + dataRowCount - 1, colHead + 1, rowHead + 1); 
    if (colTail < getColCount() && dataRowCount > 0)
      result.copyFrom(this, colTail, fromDataRow, getColCount() - 1, fromDataRow + dataRowCount - 1, colHead + 1 + dataColCount, rowHead + 1); 
    if (colHead > 0 && rowTail < getRowCount())
      result.copyFrom(this, 1, rowTail, colHead, getRowCount() - 1, 1, rowHead + 1 + dataRowCount); 
    if (dataColCount > 0 && rowTail < getRowCount())
      result.copyFrom(this, fromDataCol, rowTail, fromDataCol + dataColCount - 1, getRowCount() - 1, colHead + 1, rowHead + 1 + dataRowCount); 
    if (colTail < getColCount() && rowTail < getRowCount())
      result.copyFrom(this, colTail, rowTail, getColCount() - 1, getRowCount() - 1, colHead + 1 + dataColCount, rowHead + 1 + dataRowCount); 
    return result;
  }
  
  public int getDataRowCount() {
    int top = (this.scrollTopRow > 0) ? this.scrollTopRow : 1;
    int bottom = (this.scrollBottomRow > 0) ? this.scrollBottomRow : (getRowCount() - 1);
    int count = 0;
    for (int i = top; i <= bottom; i++) {
      if (getRowVisible(i))
        count++; 
    } 
    return count;
  }
  
  public int getDataColCount() {
    int left = (this.scrollTopCol > 0) ? this.scrollTopCol : 1;
    int right = (this.scrollBottomCol > 0) ? this.scrollBottomCol : (getColCount() - 1);
    int count = 0;
    for (int i = left; i <= right; i++) {
      if (getColVisible(i))
        count++; 
    } 
    return count;
  }
  
  public ExtendedDatas getExtDatas() {
    return this.extDatas;
  }
  
  public int getHorzOffset() {
    return this.horzOffset;
  }
  
  public void setHorzOffset(int offset) {
    this.horzOffset = offset;
  }
  
  public int getVertOffset() {
    return this.vertOffset;
  }
  
  public void setVertOffset(int offset) {
    this.vertOffset = offset;
  }
  
  public boolean getHorzCenter() {
    return this.horzCenter;
  }
  
  public void setHorzCenter(boolean value) {
    this.horzCenter = value;
  }
  
  public boolean getVertCenter() {
    return this.vertCenter;
  }
  
  public void setVertCenter(boolean value) {
    this.vertCenter = value;
  }
  
  public static com.jiuqi.grid.GridData bytesToGrid(byte[] data) {
    if (data == null)
      return null; 
    com.jiuqi.grid.GridData gd = new com.jiuqi.grid.GridData();
    try {
      MemStream memStream = new MemStream();
      memStream.writeBuffer(data, 0, data.length);
      memStream.setPosition(0L);
      gd.loadFromStream((Stream)memStream);
    } catch (Exception ex) {
      ex.printStackTrace();
      gd = null;
    } 
    return gd;
  }
  
  public static com.jiuqi.grid.GridData bytesToGrid(byte[] data, String encodeCharset) {
    if (data == null)
      return null; 
    com.jiuqi.grid.GridData gd = new com.jiuqi.grid.GridData();
    try {
      MemStream memStream = new MemStream();
      memStream.setUseEncode(true);
      memStream.setCharset(encodeCharset);
      memStream.writeBuffer(data, 0, data.length);
      memStream.setPosition(0L);
      gd.loadFromStream((Stream)memStream);
    } catch (Exception ex) {
      ex.printStackTrace();
      gd = null;
    } 
    return gd;
  }
  
  public static byte[] gridToBytes(com.jiuqi.grid.GridData grid) {
    if (grid == null)
      return null; 
    MemStream memStream = new MemStream();
    try {
      grid.saveToStream((Stream)memStream);
      return memStream.getBytes();
    } catch (StreamException ex) {
      ex.printStackTrace();
      return null;
    } 
  }
  
  public static byte[] gridToBytes(com.jiuqi.grid.GridData grid, String encodeCharset) {
    if (grid == null)
      return null; 
    MemStream memStream = new MemStream();
    try {
      memStream.setUseEncode(true);
      memStream.setCharset(encodeCharset);
      grid.saveToStream((Stream)memStream);
      return memStream.getBytes();
    } catch (StreamException ex) {
      ex.printStackTrace();
      return null;
    } 
  }
  
  public static com.jiuqi.grid.GridData base64ToGrid(String base64Str) {
    byte[] data = Base64.base64ToByteArray(base64Str);
    return bytesToGrid(data);
  }
  
  public static com.jiuqi.grid.GridData base64ToGrid(String base64Str, String encodeCharset) {
    byte[] data = Base64.base64ToByteArray(base64Str);
    return bytesToGrid(data, encodeCharset);
  }
  
  public static String gridToBase64(com.jiuqi.grid.GridData grid) {
    byte[] data = gridToBytes(grid);
    return Base64.byteArrayToBase64(data);
  }
  
  public static String gridToBase64(com.jiuqi.grid.GridData grid, String encodeCharset) {
    byte[] data = gridToBytes(grid, encodeCharset);
    return Base64.byteArrayToBase64(data);
  }
  
  public void convertData(CellConvertor convertor) {
    for (int col = 0; col < getColCount(); col++) {
      for (int row = 0; row < getRowCount(); row++) {
        CellField cf = expandCell(col, row);
        if (cf.left == col && cf.top == row) {
          GridCell cell = getCellEx(col, row);
          if (convertor.convert(this, cell))
            setCell(col, row, cell); 
        } 
      } 
    } 
  }
  
  public int getOptions() {
    return this.options;
  }
  
  public void setOptions(int value) {
    this.options = value;
    this.rects.setAutoMerge(((value & 0x1) != 0));
  }
}
